/*  equation.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include <sstream>
#include <fstream>

#include <ginac/ginac.h>

#include "integralfamily.h"
#include "sectormappings.h"
#include "int.h"
#include "functions.h"
#include "equation.h"
#include "files.h"
#include "kinematics.h"
#include "sector.h"
#include "ginacutils.h"
#include "identitygenerator.h"
#ifdef HAVE_FERMAT
#include "fermat.h"
#endif

using namespace std;
using namespace GiNaC;

namespace Reduze {

/// gives a normalized form of the expression, must recognize zero
template<class TC>
static TC normal_form(const TC& c);

#ifdef HAVE_FERMAT
template<>
GiNaC::ex normal_form(const GiNaC::ex& c) {
	if(Fermat::instance()->is_initialized()) {
		return Fermat::instance()->normalize(c);
	}
	return GiNaC::normal(c);
}
#else //HAVE_FERMAT
template<>
GiNaC::ex normal_form(const GiNaC::ex& c) {
	return GiNaC::normal(c);
}
#endif //HAVE_FERMAT
// LinCombBasic

template<class TI, class TC>
double LinCombBasic<TI, TC>::average_coeff_length() {
	typename terms_type::const_iterator it = terms.begin();
	double length = size();
	double sum = 0.0;
	while (it != terms.end()) {
		stringstream strm;
		strm << it->second;
		string str = strm.str();
		double part = str.size();
		part /= length;
		sum += part;
		++it;
	}
	return sum;
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::insert(const TI& i, const TC& e, bool normalize) {
	// looks a little complicated but needs only one map lookup
	typename std::pair<typename terms_type::iterator, bool> info =
			this->terms.insert(std::make_pair(i, e));
	typename terms_type::iterator it = info.first;
	bool inserted = info.second;
	if (!inserted) { // equation contained the integral already, add coeff
		it->second += e;
		if (normalize)
			it->second = normal_form(it->second);
	}
	/* alternatively:
	 typename terms_type::iterator it = this->terms.lower_bound(i);
	 if (it != this->terms.end() && !(this->terms.key_comp()(i, it->first)))
	 it->second += e;
	 else // new integral
	 it = this->terms.insert(it, typename terms_type::value_type(i, e));
	 */
	if (it->second == 0)
		this->terms.erase(it);
}

template<>
void LinCombBasic<INT, GiNaC::ex>::insert(const INT & i, const GiNaC::ex & e) {
	// looks a little complicated but needs only one map lookup
	std::pair<terms_type::iterator, bool> info = terms.insert(std::make_pair(i,
			e));
	terms_type::iterator it = info.first;
	bool inserted = info.second;
	if (!inserted) // equation contained the integral already, add coeff
		it->second += e;
}

template<>
void LinCombBasic<uintx, GiNaC::ex>::insert(const uintx & i,
		const GiNaC::ex & e) {
	// looks a little complicated but needs only one map lookup
	std::pair<terms_type::iterator, bool> info = terms.insert(std::make_pair(i,
			e));
	terms_type::iterator it = info.first;
	bool inserted = info.second;
	if (!inserted) // equation contained the integral already, add coeff
		it->second += e;
}

template<class TI, class TC>
void LinCombBasic<TI, TC>::insert(const TI & i, const TC & e) {
	// looks a little complicated but needs only one map lookup
	typename std::pair<typename terms_type::iterator, bool> info =
			terms.insert(std::make_pair(i, e));
	bool inserted = info.second;
	if (!inserted) // equation contained the integral already
		ABORT("Can't handle multiple occurrence of integral " << i);
}

template<class TI, class TC>
void LinCombBasic<TI, TC>::save(const string & filename, bool append) const {
	ofstream out;
	if (append)
		out.open(filename.c_str(), ios::app);
	else
		out.open(filename.c_str());
	if (!out.good())
		ABORT("Error opening file: " << filename);
	out << *this;
	out.close();
}

// LinCombBasic friends

template<class TI, class TC>
std::ostream & operator<<(std::ostream & os, const LinCombBasic<TI, TC> & lc) {
	typename LinCombBasic<TI, TC>::terms_type::const_reverse_iterator it;
	if (!lc.name_.empty())
		os << "name = " << lc.name_ << "\n";
	for (it = lc.terms.rbegin(); it != lc.terms.rend(); ++it) {
		os << (it == lc.terms.rbegin() ? "  " : "    ") //
				<< it->first << "\n" //
				<< (it == lc.terms.rbegin() ? "   " : "     ") //
				<< it->second << "\n";
	}
	os << ";";
	return os;
}

template<class TI, class TC>
bool LinCombBasic<TI, TC>::operator <(const LinCombBasic<TI, TC> & other) const {
	// compare integrals of the equations
	typename terms_type::const_reverse_iterator it, oit;
	it = terms.rbegin();
	oit = other.terms.rbegin();
	for (; it != terms.rend() && oit != other.terms.rend(); ++it, ++oit) {
		if (it->first != oit->first)
			return (it->first < oit->first);
	}
	if (it != terms.rend() || oit != other.terms.rend())
		return (it == terms.rend());
	// both equations contain the same integrals, now check coefficients
	it = terms.rbegin();
	oit = other.terms.rbegin();
	for (; it != terms.rend() && oit != other.terms.rend(); ++it, ++oit) {
		int coeffcmp = it->second.compare(oit->second);
		if (coeffcmp != 0)
			return coeffcmp;
	}
	return false;
}

/// \todo: remove this workaround by providing CoeffIndex::compare()
template<>
bool LinCombBasic<uintx, CoeffIndex>::operator <(const LinCombBasic<uintx,
		CoeffIndex> & other) const {
	// compare integrals of the equations
	terms_type::const_reverse_iterator it, oit;
	it = terms.rbegin();
	oit = other.terms.rbegin();
	for (; it != terms.rend() && oit != other.terms.rend(); ++it, ++oit) {
		if (it->first != oit->first)
			return (it->first < oit->first);
	}
	if (it != terms.rend() || oit != other.terms.rend())
		return (it == terms.rend());
	// both equations contain the same integrals, now check coefficients
	it = terms.rbegin();
	oit = other.terms.rbegin();
	for (; it != terms.rend() && oit != other.terms.rend(); ++it, ++oit) {
		if (it->second != oit->second)
			return it->second < oit->second;
	}
	return false;
}

template<class TI, class TC>
bool LinCombBasic<TI, TC>::operator ==(const LinCombBasic<TI, TC> & other) const {
	if (size() != other.size())
		return false;
	// compare the more complicated integrals first
	typename terms_type::const_reverse_iterator it = terms.rbegin();
	typename terms_type::const_reverse_iterator oit = other.terms.rbegin();
	while (it != terms.rend()) {
		if (it->first != oit->first)
			return false;
		if (!(it->second == oit->second))
			return false;
		++it;
		++oit;
	}
	return true;
}

template<class TI, class TC>
bool LinCombBasic<TI, TC>::operator !=(const LinCombBasic<TI, TC> & other) const {
	return !(*this == other);
}

// LinCombAlgebraic

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::insert(const_iterator first, const_iterator last) {
	this->terms.insert(first, last);
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::normalize() {
	typename terms_type::iterator it;
	for (it = this->terms.begin(); it != this->terms.end();) {
		it->second = normal_form(it->second);
		if (it->second == 0)
			this->terms.erase(it++);
		else
			++it;
	}
}

template<class TI, class TC>
bool LinCombAlgebraic<TI, TC>::has_symbol(const GiNaC::symbol & symb) const {
	typename terms_type::const_iterator it;
	for (it = this->terms.begin(); it != this->terms.end(); ++it)
		if (it->second.has(symb))
			return true;
	return false;
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::series(const GiNaC::symbol& param,
		const GiNaC::ex& point, int order) {
	typename terms_type::iterator t;
	for (t = this->terms.begin(); t != this->terms.end(); ++t)
		if (t->second.has(param))
			t->second = t->second.series(param == point, order + 1);
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::apply(const GiNaC::exmap & m) {
	typename terms_type::iterator it = this->terms.begin();
	while (it != this->terms.end()) {
		it->second = it->second.subs(m);
		++it;
	}
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::replace(const EqAlgebraic<TI, TC> & sol,
		bool normalize) {
	if (sol.terms.empty())
		return;
	// find leading term with non-zero coefficient
	typename terms_type::const_iterator st = --sol.terms.end();
	ex lc;
	do {
		lc = normal_form(st->second);
	} while (lc == 0 && st != sol.terms.begin() && (--st, true));
	if (lc == 0)
		return;
	// search if this linear combination contains the leading integral
	typename terms_type::iterator it_found = this->terms.find(st->first);
	if (it_found == this->terms.end())
		return;
	// replace term in this linear combination with the solution
	coeff_type prefac = coeff_type(-1) * it_found->second / lc;
	this->terms.erase(it_found);
	while (st != sol.terms.begin() && (--st, true))
		insert(st->first, st->second * prefac, normalize);
}

// these wrappers could be avoided with a dedicated 'small INT' class with uintx rep
template<class TI>
void to_mma_stream(const TI& integral, std::ostream& os) {
	integral.to_mma_stream(os);
}
template<>
void to_mma_stream<uintx> (const uintx& integral, std::ostream& os) {
	os << integral;
}
template<class TI>
void to_maple_stream(const TI& integral, std::ostream& os) {
	integral.to_maple_stream(os);
}
template<>
void to_maple_stream<uintx> (const uintx& integral, std::ostream& os) {
	os << integral;
}
template<class TI>
void to_form_stream(const TI& integral, std::ostream& os) {
	integral.to_form_stream(os);
}
template<>
void to_form_stream<uintx> (const uintx& integral, std::ostream& os) {
	os << integral;
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::to_mma_stream(ostream & out) const {
	out << this->name() << " -> ";
	if (this->empty()) {
		out << "0";
		return;
	}
	typename terms_type::const_reverse_iterator t;
	for (t = this->terms.rbegin(); t != this->terms.rend();) {
		// 6 spaces in front of the prefactor, 2 spaces in front of the integral
		out << "\n  ";
		Reduze::to_mma_stream(t->first, out);
		out << " *\n      (";
		subs_conjugate()(t->second).print(print_mma(out));
		//t->second.print(print_mma(out));
		out << ")";
		++t;
		if (t != this->terms.rend())
			out << " +";
	}
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::to_maple_stream(ostream & out) const {
	out << this->name() << " = ";
	if (this->empty()) {
		out << "0";
		return;
	}
	typename terms_type::const_reverse_iterator t;
	for (t = this->terms.rbegin(); t != this->terms.rend();) {
		// 6 spaces in front of the prefactor, 2 spaces in front of the integral
		out << "\n  ";
		Reduze::to_maple_stream(t->first, out);
		out << " *\n      (";
		t->second.print(print_maple(out));
		out << ")";
		++t;
		if (t != this->terms.rend())
			out << " +";
	}
}

template<class TI, class TC>
void LinCombAlgebraic<TI, TC>::to_form_stream(ostream & out) const {
	out << "id " << this->name() << " = ";
	if (this->empty()) {
		out << "0;\n";
		return;
	}
	typename terms_type::const_reverse_iterator t;
	for (t = this->terms.rbegin(); t != this->terms.rend(); ++t) {
		out << "\n  + ";
		Reduze::to_form_stream(t->first, out);
		out << "\n    * (";
		t->second.print(print_form(out));
		out << ")";
	}
	out << ";\n";
}

// EqAlgebraic

template<class TI, class TC>
void EqAlgebraic<TI, TC>::solve() {
	if (this->terms.empty())
		return;
	typename terms_type::iterator l = --this->terms.end();
	coeff_type div = normal_form(l->second);
	if (div == 0) {
		this->terms.erase(l);
		return solve(); // recurse
	} else {
		l->second = 1;
		for (typename terms_type::iterator t = this->terms.begin(); t != l;) {
			t->second = normal_form(t->second / div);
			if (t->second == 0)
				this->terms.erase(t++);
			else
				++t;
		}
	}
}

template<class TI, class TC>
void EqAlgebraic<TI, TC>::to_mma_stream(ostream & out) const {
	typename terms_type::const_reverse_iterator it = this->terms.rbegin();
	if (this->empty()) {
		WARNING("MMA output: Empty equation not printed");
		return;
	}
	if (this->size() == 1) {
		Reduze::to_mma_stream(it->first, out);
		out << " -> 0";
		return;
	}
	if (it->second != 1) {
		WARNING("MMA output: left hand side of the equation still has a prefactor not equal to one!");
		out << "(";
		it->second.print(print_mma(out));
		out << ") * ";
	}
	Reduze::to_mma_stream(it->first, out);
	out << " -> ";
	++it;
	while (it != this->terms.rend()) {
		// 6 spaces in front of the prefactor, 2 spaces in front of the integral
		out << "\n  ";
		Reduze::to_mma_stream(it->first, out);
		out << " *\n      (";
		(-it->second).print(print_mma(out)); // with a minus sign
		out << ")";
		++it;
		if (it != this->terms.rend())
			out << " +";
	}
}

template<class TI, class TC>
void EqAlgebraic<TI, TC>::to_maple_stream(ostream & out) const {
	typename terms_type::const_reverse_iterator it = this->terms.rbegin();
	if (this->empty()) {
		WARNING("Maple output: Empty equation not printed");
		return;
	}
	if (this->size() == 1) {
		Reduze::to_maple_stream(it->first, out);
		out << " = 0";
		return;
	}
	if (it->second != 1) {
		WARNING("Maple output: left hand side of the equation still has a prefactor not equal to one!");
		out << "(";
		it->second.print(print_maple(out));
		out << ") * ";
	}
	Reduze::to_maple_stream(it->first, out);
	out << " = ";
	++it;
	while (it != this->terms.rend()) {
		// 6 spaces in front of the prefactor, 2 spaces in front of the integral
		out << "\n  ";
		Reduze::to_maple_stream(it->first, out);
		out << " * \n      (";
		(-it->second).print(print_maple(out)); // with a minus sign
		out << ")";
		++it;
		if (it != this->terms.rend())
			out << " +";
	}
}

template<class TI, class TC>
void EqAlgebraic<TI, TC>::to_form_stream(ostream & out) const {
	typename terms_type::const_reverse_iterator it = this->terms.rbegin();
	if (this->empty()) {
		WARNING("FORM output: Empty equation not printed");
		return;
	}
	if (this->size() == 1) {
		out << "id ";
		Reduze::to_form_stream(it->first, out);
		out << " = 0;\n";
		return;
	}
	if (it->second != 1) {
		WARNING("FORM output: left hand side of the equation still has a prefactor not equal to one!");
		out << "id (";
		it->second.print(print_form(out));
		out << ") * ";
		Reduze::to_form_stream(it->first, out);
	} else {
		out << "id ";
		Reduze::to_form_stream(it->first, out);
		out << " = ";
	}
	++it;
	while (it != this->terms.rend()) {
		out << "\n  + ";
		Reduze::to_form_stream(it->first, out);
		out << "\n    * (";
		(-it->second).print(print_form(out)); // with a minus sign
		out << ")";
		++it;
	}
	out << ";\n";
}

// helper templates

namespace {
/// returns value for given key, aborts if not found
template<class TKey, class TVal>
TVal translate(const std::map<TKey, TVal>& m, const TKey& k) {
	typename map<TKey, TVal>::const_iterator it = m.find(k);
	if (it == m.end())
		ABORT("no mapping for integral " << k << " found");
	return it->second;
}
}

// EquationLight

EquationLight::EquationLight(const Equation& eq) {
	LinCombBasic<uintx, GiNaC::ex>::terms_type::const_iterator it;
	for (it = eq.terms.begin(); it != eq.terms.end(); ++it) {
		stringstream ss;
		ss << it->second;
		insert(it->first, ss.str());
	}
}

EquationLight::EquationLight(const Identity& eq, const std::map<INT, uintx>& m) {
	LinCombBasic<INT, GiNaC::ex>::terms_type::const_iterator it;
	for (it = eq.terms.begin(); it != eq.terms.end(); ++it) {
		stringstream ss;
		ss << it->second;
		insert(translate(m, it->first), ss.str());
	}
}

// EquationHLight

EquationHLight::EquationHLight(const LinCombBasic<INT, string>& lc) {
	LinCombBasic<INT, string>::terms_type::const_iterator t;
	for (t = lc.begin(); t != lc.end(); ++t)
		insert(t->first, t->second);
	set_name("");
}

// Equation

Equation::Equation(const EquationLight& eq) {
	lst syms = Files::instance()->kinematics()->kinematic_invariants();
	syms.append(Files::instance()->kinematics()->dimension());
	LinCombBasic<uintx, std::string>::terms_type::const_iterator it;
	for (it = eq.terms.begin(); it != eq.terms.end(); ++it) {
		const int i = it->first;
		GiNaC::ex coeff(it->second, syms);
		insert(i, coeff, false);
	}
}

Equation::Equation(const Identity& eq, const std::map<INT, uintx>& m) {
	LinCombBasic<INT, GiNaC::ex>::terms_type::const_iterator it;
	for (it = eq.terms.begin(); it != eq.terms.end(); ++it)
		insert(translate(m, it->first), it->second, false);
}

// Identity

Identity::Identity(const EquationHLight& eq) {
	LinCombBasic<INT, std::string>::terms_type::const_iterator t;
	for (t = eq.terms.begin(); t != eq.terms.end(); ++t) {
		lst s = t->first.integralfamily()->kinematics()->kinematic_invariants();
		s.append(t->first.integralfamily()->kinematics()->dimension());
		insert(t->first, GiNaC::ex(t->second, s));
	}
}

Identity::Identity(const EquationLight& eq, const std::map<uintx, INT>& m) {
	LinCombBasic<uintx, std::string>::terms_type::const_iterator t;
	for (t = eq.terms.begin(); t != eq.terms.end(); ++t) {
		INT integ(translate(m, t->first));
		lst s = integ.integralfamily()->kinematics()->kinematic_invariants();
		s.append(integ.integralfamily()->kinematics()->dimension());
		GiNaC::ex coeff(t->second, s);
		insert(integ, coeff);
	}
}

Identity::Identity(const Equation& eq, const std::map<uintx, INT>& m) {
	LinCombBasic<uintx, GiNaC::ex>::terms_type::const_iterator t;
	for (t = eq.terms.begin(); t != eq.terms.end(); ++t)
		insert(translate(m, t->first), t->second);
}

Identity::Identity(const LinCombAlgebraic<INT, GiNaC::ex>& lc) {
	LinCombBasic<INT, GiNaC::ex>::terms_type::const_iterator t;
	for (t = lc.terms.begin(); t != lc.terms.end(); ++t)
		insert(t->first, t->second, false);
	name_.clear();
}

void Identity::toggle_metric_convention() {
	Identity id;
	terms_type::iterator t = terms.begin();
	while (t != terms.end()) {
		INT i(t->first);
		i.set_metric_mostly_plus(!t->first.uses_metric_mostly_plus());
		int r = static_cast<int> (t->first.r());
		int s = static_cast<int> (t->first.s());
		GiNaC::ex c = (t->second) * pow(-1, r + s);
		id.insert(i, c, false);
		++t;
	}
	*this = id;
}

/*

 // decomposes terms in a string "int1\ncoeff1\nint2\ncoeff2\n;\n"
 list<pair<string, string> > split_terms_string(const std::string& s) {
 list<pair<string, string> > res;
 size_t from = 0;
 while (true) {
 size_t to = s.find("\n", from);
 string sub1 = s.substr(from, to == string::npos ? to : to - from);
 if (sub1 == ";")
 return res;
 if (to == string::npos || to + 1 >= s.size()) // assert valid next 'from'
 throw runtime_error("incomplete terms in string");
 from = to + 1;
 to = s.find("\n", from);
 sub2 = s.substr(from, to == string::npos ? to : to - from);
 res.insert(make_pair(sub1, sub2));
 if (to == string::npos || to + 1 >= s.size()) // assert valid next 'from'
 throw runtime_error("incomplete terms in string");
 from = to + 1;
 }
 }
 // extracts key/value pairs in string "key1 = val1\nkey2 = val2\nno_more_eq_sign"
 // also returns position at which residual string without '=' sign starts
 map<string, string> extract_settings(const std::string& s, size_t& pos) {
 map<string, string> result;
 while (pos < s.size()) {
 size_t to = s.find("\n", pos);
 string sub = s.substr(pos, to == string::npos ? to : to - pos);
 size_t eq = sub.find("=");
 if (eq == string::npos)
 return result;
 string key = chop_leading_trailing_whitespaces(sub.substr(from, eq));
 string val = chop_leading_trailing_whitespaces(sub.substr(eq + 1, to));
 result[key] = val;
 pos = to + 1;
 }
 pos = string::npos;
 return result;
 }
 */

LinearCombination::LinearCombination(istream& is, const GiNaC::lst& symbols) {
	string l, r;
	getline(is, l);
	if (l.find("=") != string::npos) { // scan leading name = NAME line
		ICharArrayBuffer lb(l.c_str(), l.size());
		istream ls(&lb);
		string s1, s2, s3;
		ls >> s1; // name
		ls >> s2; // =
		ls >> s3; // NAME
		if (s1 != "name" || s2 != "=")
			throw runtime_error("missing name in LinCombHLight string");
		set_name(s3);
		getline(is, l);
	}
	while (is && chop_whitespaces(l) != ";" && getline(is, r)) {
		insert(INT(l), GiNaC::ex(r, symbols));
		getline(is, l);
	}
	if (!is)
		throw runtime_error("syntax error in LinearCombinations string");
}

LinCombHLight::LinCombHLight(istream& is) {
	string l, r;
	getline(is, l);
	if (l.find("=") != string::npos) { // scan leading name = NAME line
		ICharArrayBuffer lb(l.c_str(), l.size());
		istream ls(&lb);
		string s1, s2, s3;
		ls >> s1; // name
		ls >> s2; // =
		ls >> s3; // NAME
		if (s1 != "name" || s2 != "=")
			throw runtime_error("missing name in LinCombHLight string");
		set_name(s3);
		getline(is, l);
	}
	while (is && chop_whitespaces(l) != ";" && getline(is, r)) {
		insert(INT(l), chop_whitespaces(r));
		getline(is, l);
	}
	if (!is)
		throw runtime_error("syntax error in LinCombHLight string");
}

void LinearCombination::toggle_metric_convention() {
	LinearCombination lc;
	terms_type::iterator t = terms.begin();
	while (t != terms.end()) {
		INT i(t->first);
		i.set_metric_mostly_plus(!t->first.uses_metric_mostly_plus());
		int r = static_cast<int> (t->first.r());
		int s = static_cast<int> (t->first.s());
		GiNaC::ex c = (t->second) * pow(-1, r + s);
		lc.insert(i, c, false);
		++t;
	}
	*this = lc;
}

void LinearCombination::apply_shifts(bool normalize) {
	set<INT> ints;
	find_INT(ints);
	for (set<INT>::const_iterator i = ints.begin(); i != ints.end(); ++i) {
		Identity res = IdentityGenerator::get_shift_identity(*i);
		if (!res.empty())
			replace(res, normalize);
	}
}

void Identity::cut() {
	terms_type::iterator it = terms.begin(), tmpit;
	while (it != terms.end()) {
		if (it->first.is_cut_to_zero()) {
			tmpit = it++;
			terms.erase(tmpit);
		} else {
			++it;
		}
	}
}

void Identity::reconstruct_symbol_replaced_by_one() {
	using namespace GiNaC;
	const Kinematics* kin = Files::instance()->kinematics();
	const symbol* norm = kin->symbol_to_replace_by_one();
	if (norm == 0) {
		LOGX("No symbol found which was set to one and needs reconstruction");
		return;
	} else if (has_symbol(*norm)) {
		ABORT("Symbol " << *norm << " to reconstruct already present\n" << *this);
	}
	ex dim_norm = kin->find_mass_dimension(*norm);
	exmap recover_invars;
	lst invars = kin->kinematic_invariants();
	for (lst::const_iterator s = invars.begin(); s != invars.end(); ++s) {
		if (!is_a<symbol> (*s))
			ABORT("This is no symbol: " << *s);
		ex dim_s = kin->find_mass_dimension(ex_to<symbol> (*s));
		recover_invars[*s] = (*s) * pow(*norm, -dim_s / dim_norm);
	}
	terms_type::iterator t;
	for (t = terms.begin(); t != terms.end(); ++t) {
		const INT& i = t->first;
		ex& c = t->second;
		ex dim_t_rel = pow(*norm, 2 * (i.s() - i.r()) / dim_norm);
		c = normal(c.subs(recover_invars) / dim_t_rel);
	}
}

Identity Identity::replace_family_and_coefficients(
		const IntegralFamily* old_ic, const IntegralFamily* new_ic,
		const GiNaC::exmap& m) const {
	Identity id;
	for (const_iterator it = begin(); it != end(); ++it) {
		ASSERT(it->first.integralfamily() == old_ic);
		id.insert(INT(new_ic, it->first.v()), it->second.subs(m));
	}
	return id;
}

// global functions

// NEEDS an EXPANDED expression,
void propagator_products_to_integrals(const ex& e, LinCombAlgebraic<INT,
		GiNaC::ex>& lincomb, const IntegralFamily* ic) {
	if (is_a<add> (e)) {
		int num = e.nops();
		for (int i = 0; i < num; ++i) {
			INT integral(ic);
			ex coeff;
			if (!ic->match_integrand(e.op(i), integral, coeff)) {
				stringstream ss;
				ss << "can't match to propagators of family " << ic->name()
						<< ": " << e.op(i);
				throw runtime_error(ss.str());
			}
			lincomb.insert(integral, coeff, false);
		}
	} else {
		INT integral(ic);
		ex coeff;
		if (!ic->match_integrand(e, integral, coeff)) {
			stringstream ss;
			ss << "can't match to propagators of family " << ic->name() << ": "
					<< e;
			throw runtime_error(ss.str());
		}
		lincomb.insert(integral, coeff, false);
	}
}

// explicit instantiations
template class LinCombBasic<uintx, std::string> ;
template class LinCombBasic<INT, std::string> ;
template class LinCombBasic<uintx, CoeffIndex> ;
template class LinCombBasic<uintx, GiNaC::ex> ;
template class LinCombBasic<INT, GiNaC::ex> ;
template class LinCombAlgebraic<uintx, GiNaC::ex> ;
template class LinCombAlgebraic<INT, GiNaC::ex> ;
template class EqAlgebraic<uintx, GiNaC::ex> ;
template class EqAlgebraic<INT, GiNaC::ex> ;

} // namespace Reduze
