/*  equation2.cpp
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */


/** @file equation2.cpp
 *
 *  Copyright (C) 2009-2010  Cedric Studerus
 *  Copyright (C) 2010       Andreas von Manteuffel
 *  This file is part of the package Reduze.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#include <ginac/ginac.h>
#include "integralfamily.h"
#include "functions.h"
//#include "functions_yaml.h"
#include "equation2.h"
#include "ginacutils.h"
#include "propagator.h"

using namespace std;
using namespace GiNaC;

namespace Reduze {

// OP


ostream & operator<<(ostream & strm, const OP & in) {
	strm << "OP[" << in.n_ << "]";
	return strm;
}

// OPPROD

OPPROD::OPPROD(const GiNaC::ex & e, const OP & op) :
	fac_(e), prod_(1, op) {
	ex pattern1 = Propagator(GiNaC::wild(1), GiNaC::wild(2));
	ex pattern2 = Propagator(GiNaC::wild(1), GiNaC::wild(2), GiNaC::wild(3));
	if (fac_.has(pattern1) || fac_.has(pattern2)) {
		std::stringstream ss;
		ss << "coefficient " << fac_ << " contains propagator";
		throw std::runtime_error(ss.str());
	}
}

void OPPROD::unique() {
	if (size() < 2)
		return;
	prod_.sort();
	list<OP>::iterator it1 = prod_.begin(), it2;
	while (it1 != prod_.end()) {
		if (it1->n_ == 0) {
			it1 = prod_.erase(it1);
		} else {
			it2 = it1;
			++it2;
			while (it2 != prod_.end()) {
				if (it1->n_ == -it2->n_) {
					it2 = prod_.erase(it2);
					it1 = prod_.erase(it1);
					it2 = it1;
					if (it2 == prod_.end()) {
						if (size() == 0) {
							*this = OPPROD(fac_, OP()); // product should have at least an unity
						}
						return;
					} else {
						++it2;
					}
				} else {
					++it2;
				}
			}
			// it1 is not at the end
			++it1;
		}
	}
}

// ordering not complete
bool OPPROD::operator<(const OPPROD & other) const {
	if (size() < other.size())
		return true;
	return false;
}

bool OPPROD::operator==(const OPPROD & other) const {
	if (*this < other || other < *this)
		return false;
	list<OP>::const_iterator it1 = prod_.begin();
	list<OP>::const_iterator it2 = other.prod_.begin();
	while (it1 != prod_.end()) {
		if (*it1 != *it2)
			return false;
		++it1;
		++it2;
	}
	return true;
}

bool OPPROD::operator!=(const OPPROD & other) const {
	return !(*this == other);
}

OPPROD & OPPROD::operator*=(const OPPROD & other) {
	prod_.insert(prod_.begin(), other.prod_.begin(), other.prod_.end());
	fac_ *= other.fac_;
	fac_.normal();
	unique();
	return *this;
}

const OPPROD operator*(const OPPROD & a, const OPPROD & b) {
	OPPROD tmp = a;
	tmp *= b;
	return tmp;
}

ostream & operator<<(ostream & strm, const OPPROD & in) {
	strm << "(" << in.fac_ << ") * ";
	list<OP>::const_iterator it = in.prod_.begin();
	while (it != in.prod_.end()) {
		strm << *it;
		++it;
		if (it != in.prod_.end()) {
			strm << ".";
		}
	}
	return strm;
}

// OPSUM


void OPSUM::unique() {
	list<OPPROD>::iterator it = sum.begin();
	while (it != sum.end()) {
		it->unique();
		++it;
	}
}

void OPSUM::collect() {
	if (size() < 2)
		return;
	list<OPPROD>::iterator it1 = sum.begin(), it2;
	bool bu;
	while (it1 != sum.end()) {
		if (!it1->fac_.match(0)) {
			it2 = it1;
			++it2;
			bu = false;
			while (it2 != sum.end()) {
				if (*it2 == *it1) {
					it1->fac_ += it2->fac_;
					it2 = sum.erase(it2);
					bu = true;
				} else {
					++it2;
				}
			} // it2
			if (bu) {
				it1->fac_ = normal(it1->fac_);
				if ((it1->fac_).match(0))
					it1 = sum.erase(it1);
				if (it1 == sum.end())
					return;
			} else {
				++it1;
			}
		} else {
			it1 = sum.erase(it1);
		}
	}
}

OPSUM & OPSUM::operator+=(const OPSUM & other) {
	sum.insert(sum.end(), other.sum.begin(), other.sum.end());
	unique();
	collect();
	return *this;
}

OPSUM & OPSUM::operator-=(const OPSUM & other) {
	OPSUM tmp = other;
	list<OPPROD>::iterator it = tmp.sum.begin();
	while (it != tmp.sum.end()) {
		it->fac_ = -it->fac_;
		++it;
	}
	sum.insert(sum.end(), tmp.sum.begin(), tmp.sum.end());
	unique();
	collect();
	return *this;
}

OPSUM & OPSUM::operator*=(const OPSUM & other) {
	OPSUM neu = *this;
	sum.clear();
	list<OPPROD>::const_iterator it1 = neu.sum.begin(), it2;
	while (it1 != neu.sum.end()) {
		it2 = other.sum.begin();
		while (it2 != other.sum.end()) {
			sum.push_back(*it1 * *it2);
			++it2;
		}
		++it1;
	}
	unique();
	collect();
	return *this;
}

const OPSUM operator+(const OPSUM & a, const OPSUM & b) {
	OPSUM tmp = a;
	tmp += b;
	return tmp;
}

const OPSUM operator-(const OPSUM & a, const OPSUM & b) {
	OPSUM tmp = a;
	tmp -= b;
	return tmp;
}

const OPSUM operator*(const OPSUM & a, const OPSUM & b) {
	OPSUM tmp = a;
	tmp *= b;
	return tmp;
}

ostream & operator<<(ostream & strm, const OPSUM & in) {
	list<OPPROD>::const_iterator it = in.sum.begin();
	if (it == in.sum.end()) {
		strm << "0";
	}
	while (it != in.sum.end()) {
		strm << *it;
		++it;
		if (it != in.sum.end())
			strm << " + ";
	}
	return strm;
}

// class INT2 members


INT2::INT2(const IntegralFamily* i) :
	fac_(1), integralfamily_(i) {
	ASSERT(integralfamily_ != 0);
	ep_ = GiNaC::exvector(i->propagator_exponents());
}

INT2::INT2(const INT2 & in) :
	fac_(in.fac_), ep_(in.ep_), integralfamily_(in.integralfamily()) {
}

INT2::INT2(const IntegralFamily* i, const GiNaC::exvector& v,
		const GiNaC::ex& fac) :
	fac_(fac), ep_(v), integralfamily_(i) {
	if (!freeof(fac, i->loop_momenta()))
		ABORT("Coefficient " << fac << " contains loop momentum: " << i->loop_momenta());
}

INT2 & INT2::operate(const OP & op) {
	int n = op.n();
	ASSERT(static_cast<size_t>(abs(n)) <= ep_.size());
	if (n > 0)
		++ep_[n - 1];
	if (n < 0)
		--ep_[-n - 1];
	return *this;
}

INT2 & INT2::operate(const OPPROD & prod) {
	list<OP>::const_iterator it = prod.prod_.begin();
	while (it != prod.prod_.end()) {
		operate(*it);
		++it;
	}
	if (!freeof(prod.fac(), integralfamily_->loop_momenta())) {
		stringstream ss;
		ss << "coefficient " << prod.fac() << " contains loop momentum: "
				<< integralfamily_->loop_momenta();
		throw runtime_error(ss.str());
	}
	//	if (!freeof(prod.get_fac(), integralfamily_->loop_momenta()))
	//		ABORT("Coefficient " << prod.get_fac() << " contains loop momentum: " << integralfamily_->loop_momenta());
	fac_ *= prod.fac();
	return *this;
}

void INT2::operate(const OPSUM & sum, Equation2 & eq) const {
	list<OPPROD>::const_iterator it = sum.sum.begin();
	while (it != sum.sum.end()) {
		INT2 tmp(*this);
		tmp.operate(*it);
		eq.append(tmp);
		++it;
	}
}

ostream & operator<<(ostream & strm, const INT2 & in) {
	strm << "(" << in.fac_ << ") * INT" << in.ep_;
	return strm;
}

void Equation2::operate(const OPSUM & sum) {
	list<INT2>::const_iterator it = l.begin();
	while (it != l.end()) {
		Equation2 tmp;
		it->operate(sum, tmp);
		l.splice(l.end(), tmp.l);
		++it;
	}
}

LinearCombination Equation2::apply(const INT& in) const {
	const IntegralFamily* ic = in.integralfamily();
	// map to replace symbols by numbers;
	exmap m;
	exvector a = ic->propagator_exponents();
	for (size_t j = 0; j < ic->num_propagators(); j++)
		m[a[j]] = static_cast<int> (in.v_i(j));

	// apply the map 'm' and convert to LinearCombination
	LinearCombination result;
	for (list<INT2>::const_iterator i = l.begin(); i != l.end(); ++i) {
		if (i->integralfamily() != ic)
			ABORT("integral family mismatch");
		ex fac = i->fac().subs(m);
		exvector ev = i->v();
		vector<int8> v;
		v.reserve(ev.size());
		for (exvector::const_iterator vi = ev.begin(); vi != ev.end(); ++vi)
			v.push_back(
					static_cast<int8> (ex_to<numeric> (vi->subs(m)).to_int()));
		result.insert(INT(ic, v), fac);
	}
	return result;
}

ostream & operator<<(std::ostream & strm, const Equation2 & in) {
	list<INT2>::const_iterator it = in.l.begin();
	while (it != in.l.end()) {
		strm << *it;
		++it;
		if (it != in.l.end())
			strm << " + ";
	}
	return strm;
}

// EquationList2

void EquationList2::save(const std::string & filename) {
	ofstream out;
	out.open(filename.c_str());
	if (!(out.good())) {
		cerr << "EquationList2::save(...) : error opening file: " << filename
				<< endl;
		abort();
	}
	out << *this;
	out.close();
}

ostream & operator<<(ostream & strm, const EquationList2 & in) {
	list<Equation2>::const_iterator it = in.l.begin();
	strm << "{";
	while (it != in.l.end()) {
		strm << *it;
		++it;
		if (it != in.l.end())
			strm << "," << endl;
	}
	strm << "}";
	return strm;
}

} // namespace Reduze


