/*  sectormappings.h
 *
 *  Copyright (C) 2010-2012 Andreas von Manteuffel
 *  Copyright (C) 2010-2012 Cedric Studerus
 *
 *  This file is part of the package Reduze 2.
 *  It is distributed under the GNU General Public License version 3
 *  (see the file GPL-3.0.txt or http://www.gnu.org/licenses/gpl-3.0.txt).
 */

#ifndef SECTORMAPPINGS_H_
#define SECTORMAPPINGS_H_

#include "integralfamily.h"
#include <set>
#include <map>
#include "sector.h"
#include "graph.h"
#include "functions.h"

// forward declarations
namespace YAML {
class Node;
class Emitter;
}

namespace Reduze {

/// zero sectors, sector relations and sector symmetries
/** one-to-one mappings based on the permutation symmetries of IntegralFamily
 ** are not stored explicitly
 **  */
class SectorMappings {
public:

	// constructors

	/// constructs empty sector mappings
	SectorMappings(const IntegralFamily* ic);

	/// constructs sector mappings, copying zero sectors from source integral family
	SectorMappings(const CrossedIntegralFamily* ic);

	/// sets up the sector mappings, updates the shift targets
	/**
	 * determines the graphs of the sectors
	 * finds shifts to sectors in 'shift_targets' (must be sectors from lower integral families) and to itself
	 * finds zero sectors
	 *
	 * if sector_shifts is false then no graphs are constructed, no shifts are found and
	 * only zero sectors are determined
	 */
	void setup(std::set<SectorGL>& shift_targets,
			std::set<SectorGL>& local_shift_targets,
			bool find_sector_relations, bool find_sector_symmetries,
			bool find_zero_sectors, bool minimize_by_twists,
			bool construct_minimal_graphs, bool minimize_target_crossings);

	// getters

	inline const IntegralFamily* integralfamily() const {
		return integralfamily_;
	}
	inline const std::set<int>& zero_sectors() const {
		return zero_sectors_;
	}
	inline const std::set<int>& sectors_without_graph() const {
		return sectors_without_graph_;
	}
	inline const std::map<int, Sector>& sector_relations() const {
		return sector_relations_;
	}
	inline const std::map<int, GiNaC::exmap>& sector_relation_shifts() const {
		return sector_relation_shifts_;
	}
	inline const std::map<int, std::list<GiNaC::exmap> >& sector_symmetries() const {
		return sector_symmetries_;
	}

	static bool is_in_sectors_without_graph(const Sector& sector);
	bool is_in_sectors_without_graph_nonstatic(int sector_id) const;

	static bool is_in_zero_sectors(const Sector& sector);
	bool is_in_zero_sectors_non_static(int sector_id) const;

	/**
	 *  returns all sectors of this family which are minimal w.r.t global permutation
	 *  symmetries and for which no shift to any lower sector exists */
	void find_shift_targets(std::set<Sector>&) const;

	/**
	 *  returns all sectors of this family which are minimal w.r.t global permutation
	 *  symmetries and for which no shift to any lower sector exists and which are not
	 *  marked as unphysical (as sectors without a graph)
	 *  The returned sectors might be zeros sectors */
	void find_shift_targets_with_graph(std::set<SectorGL>& shift_targets,
			bool minimize_by_twists, bool construct_minimal_graphs) const;

	/* terminology:
	 * The permutation symmetries of the integral family define for each sector
	 * a class of equivalent sectors. The minimal sector (with minimal sector id)
	 * of such a class is called the "minimal equivalent" or just "equivalent"
	 *
	 * If shift relations are defined then we call the sector to which a sector
	 * can be shifted the "shift targed sector" of just the "shifted".
	 * The shifted sector is not automatically the "minimal equivalent"
	 *
	 * Sectors, which cannot be modified by perm. symmetries or shifts
	 * anymore, we call reduced sectors
	 */

	/// returns the minimal equivalent integral of integral 'integral'
	/** employs the permutation symmetries and zero sectors */
	static INT get_equivalent(const INT& integral);
	/// returns the minimal equivalent sector of sector 'sector_id'
	/** employs the permutation symmetries and zero sectors **/
	int get_equivalent(int sector_id) const;
	/// returns the minimal equivalent sector of sector 's'
	/** employs the permutation symmetries and zero sectors **/
	static Sector get_equivalent(const Sector& s);
	/// returns the shifted sector of the minimal equivalent sector of sector 'id'
	/** looks up the shift relations in sector_relations_ */
	Sector get_shifted_of_equivalent(int id) const;
	/// returns the shifted sector of the minimal equivalent sector of sector 'id'
	/** looks up the shift relations in sector_relations_ */
	static Sector get_shifted_of_equivalent(const Sector& s);
	/// returns the minimal equivalent of the result of get_shifted_of_equivalent(const Sector& s)
	static Sector get_equivalent_of_shifted_of_equivalent(const Sector& s);
	/// returns the minimal equivalent sector after recursively looking for minimal equivalents and shifted sectors of sector 'sector'
	static Sector get_reduced_sector(const Sector& s);
	/// returns the reduced sub-sectors of the sector "sector"
	/** considers the minimal equivalent and shifted sector instead of "sector" */
	static std::set<Sector> get_reduced_subsectors(const Sector& s,
			bool recurse);

	// shifts

	/// inserts shift non static
	void insert_shift_nonstatic(int from_id, const Sector& to,
			const GiNaC::exmap& shift);
	/// inserts shift static
	static void insert_shift(const Sector& src, const Sector& dest,
			const GiNaC::exmap& shift);
	/// inserts symmetry shift (maps sector onto itself)
	void insert_symmetry_shifts_nonstatic(int sector_id, const std::list<
			GiNaC::exmap>& shifts);
	/// inserts symmetry shift (maps sector onto itself)
	static void insert_symmetry_shifts(const Sector& sec, const std::list<
			GiNaC::exmap>& shifts);
	/// tests if the shift transforms the corner integral from sector 'from' to 'to' (neglecting sub-sectors)
	static bool test_shift(const Sector& from, const Sector& to,
			const GiNaC::exmap& shift);

	// graphs

	/// get the graphs of the sectors (this function is only efficient if the sectors do have a graph or are zero)
	void get_graphs(std::map<Sector, Graph>& graph_by_sector, const std::set<
			Sector>& sectors, bool minimize_by_twists,
			bool construct_minimal_graphs) const;
	static void find_canonical_graph_labels(std::map<Sector, std::pair<
			GraphLabel, GraphRelabeling> >& labels, //
			const std::map<Sector, Graph>& sectors);

	//	/// helper to find substitutions needed to implement a shift relation
	//	/** permutes integral to equivalent form such that shift can be applied **/
	//	GiNaC::exmap find_shift_for_sector_relation(INT& integral) const;


	// zero sectors
	/// insert zero sectors directly
	void insert_zero_sectors(const std::set<int>& sector_ids);
	/// test whether given sectors (keys in map) are zero and add them together with the values (sectors in the list) to the zero_sectors_
	// all sectors must belong to this integral family
	int find_zero_sectors(
			const std::map<Sector, std::list<Sector> >& sectors_by_rep,
			bool allow_longtest);

	// output

	static void print_sectortree_dot(const std::set<Sector>& sectors,
			const std::string& filename);

	friend void operator >>(const YAML::Node&, SectorMappings&);
	friend YAML::Emitter & operator<<(YAML::Emitter&, const SectorMappings&);

private:
	/// returns the minimal equivalent integral of integral 'integral'
	/** employs the permutation symmetries and zero sectors,
	 ** (pass argument by value and not by const ref to make bind1st work) */
	// \todo remove this function
	INT get_equivalent_nonstatic(INT integral) const;

	/// searches the sectors for graphs
	/** sectors must belong to SectorMappings integral family
	 * adds sectors with vanishing U-polynomial to zero_sectors_
	 * adds sectors for which no graph has been found to sectors_without_graph_ */
	void find_graphs(std::map<Sector, Graph>& graph_by_sector, const std::set<
			Sector>& sectors, bool minimize_by_twists,
			bool construct_minimal_graphs);

	// Baustelle
	// void find_graphs_from_top(std::map<Sector, Graph>& graph_by_sector,
	//		const std::set<Sector>& sectors, bool minimize_by_twists);

	/// finds shifts of the sectors in 'sectors' to the sectors in 'minimal' and to 'sectors' itself
	/**
	 * All sectors in minimal are assumed to be lower than the sectors in 'sector'.
	 * New minimal equivalent sectors of 'sectors' are added to 'minimal' and the map
	 * 'new_targets_with_equiv' is set to contain the pairs of all minimal new target from 'sectors'
	 * with a list of equivalent sectors  from 'sectors'.
	 * if update is true, new shifts for sectors for which a shift already exists in the sector mappings
	 *  are only added if they map to lower sectors. */
	void find_sector_shifts(
	/**/std::set<SectorGL>& minimal, const std::set<SectorGL>& sectors,
			std::map<SectorGL, std::list<SectorGL> >& new_targets_with_equiv,
			bool minimize_target_crossings);

	/// find all symmetry shifts (automorphism group and multi-edge permutations)
	/** returned shifts have |det| = 1 and the crossings are equivalent to the identity crossing */
	void find_symmetry_shifts(const std::set<SectorGL>& sectors);

	/// performs a small reduction to test if corner integrals reduce to zero
	/** returns true if corner integrals are found to reduce to zero */
	bool search_corner_reduction_to_zero(int sector_id, bool allow_longtest =
			true) const;

private:
	/// the integral family
	const IntegralFamily* integralfamily_;
	/// zero sector ids
	std::set<int> zero_sectors_;
	/// sectors for which no graph can be drawn
	/** sectors which are trivially zero b.c. they don't have enough independent
	 * propagators are not included */
	std::set<int> sectors_without_graph_;
	/// linear sector relations
	/** source sector id -> target sector **/
	std::map<int, Sector> sector_relations_;
	/// shifts needed to actually employ sector_relations_
	/** source sector id  -> shift to apply to source sector */
	std::map<int, GiNaC::exmap> sector_relation_shifts_;
	/// map of sector ids which are invariant under the shifts
	std::map<int, std::list<GiNaC::exmap> > sector_symmetries_;
};

} // namespace Reduze

#endif /* SECTORMAPPINGS_H_ */
